@@ -12,26 +12,32 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
 # Copyright 2018 SerialLab Corp.  All rights reserved.
+from copy import copy
+from enum import Enum
+from urllib.parse import urlparse
+
 import io
 import requests
-from copy import copy
-from urllib.parse import urlparse
-from enum import Enum
-from django.utils.translation import gettext as _
+import time
 from django.core.files.base import File
+from django.utils.translation import gettext as _
+
 from EPCPyYes.core.v1_2 import events
+from EPCPyYes.core.v1_2.events import Action
 from EPCPyYes.core.v1_2 import template_events
-from quartet_capture.rules import RuleContext
-from quartet_output import errors
-from quartet_output.transport.http import HttpTransportMixin
-from quartet_output.transport.sftp import SftpTransportMixin
-from quartet_output.models import EPCISOutputCriteria, EndPoint
-from quartet_output.parsing import SimpleOutputParser, BusinessOutputParser
+from EPCPyYes.core.v1_2.CBV import BusinessSteps, Disposition
 from quartet_capture import models, rules, errors as capture_errors
+from quartet_capture.rules import RuleContext
 from quartet_capture.tasks import create_and_queue_task
-from quartet_epcis.parsing.steps import EPCISParsingStep
 from quartet_epcis.db_api.queries import EPCISDBProxy, EntryList
 from quartet_epcis.models.choices import EventTypeChoicesEnum
+from quartet_epcis.parsing.steps import EPCISParsingStep
+from quartet_output import errors
+from quartet_output.models import EPCISOutputCriteria, EndPoint
+from quartet_output.parsing import SimpleOutputParser, BusinessOutputParser
+from quartet_output.transport.http import HttpTransportMixin
+from quartet_output.transport.sftp import SftpTransportMixin
+from quartet_output.transport.mail import MailMixin
 
 
 class ContextKeys(Enum):
@@ -285,7 +291,7 @@ class UnpackHierarchyStep(rules.Step, FilteredEventStepMixin):
                     epcs.append(epcis_event.parent_id)
                 except AttributeError:
                     pass
-                
+
         # use the db proxy to get the EPCPyYes aggregation event history back
         agg_events = self.db_proxy.get_aggregation_events_by_epcs(epcs)
         agg_events = self.process_events(agg_events)
@@ -328,8 +334,13 @@ class AddCommissioningDataStep(rules.Step, FilteredEventStepMixin):
         self.rule_context = rule_context
         self.info('%s filtered events have been found. Processing',
                   len(epcis_events))
+        all_events = set({})
         for epcis_event in epcis_events:
-            self.process_event(epcis_event, rule_context)
+            all_events = all_events.union(
+                self.process_event(epcis_event, rule_context))
+
+        rule_context.context[ContextKeys.OBJECT_EVENTS_KEY.value] = list(
+            all_events)
         self.info('Processing complete.')
 
     def process_event(self, epcis_event: events.EPCISBusinessEvent,
@@ -377,7 +388,7 @@ class AddCommissioningDataStep(rules.Step, FilteredEventStepMixin):
         all_events = self.process_events(all_events)
         self.info('Adding %s Object events to the rule context.',
                   len(all_events))
-        rule_context.context[ContextKeys.OBJECT_EVENTS_KEY.value] = all_events
+        return set(all_events)
 
     def handle_parent_entries(self, parent_entries: EntryList):
         '''
@@ -405,6 +416,7 @@ class AddCommissioningDataStep(rules.Step, FilteredEventStepMixin):
     def declared_parameters(self):
         return super().declared_parameters()
 
+
 class AddEventsByMessageStep(rules.Step, FilteredEventStepMixin):
     """
     This step will look at the rule context FILTERED_EVENTS_KEY for any filtered
@@ -413,6 +425,7 @@ class AddEventsByMessageStep(rules.Step, FilteredEventStepMixin):
     that message.  This is typically used to forward on all inbound event data
     after it has been parsed.
     """
+
     def execute(self, data, rule_context: RuleContext):
         '''
         Looks for any filtered events and then creates any object events
@@ -438,7 +451,7 @@ class AddEventsByMessageStep(rules.Step, FilteredEventStepMixin):
         Processes any filtered events.
         :param epcis_event: A filtered event.
         :param rule_context: The rule context for this step.
-        :return: None
+        :return: None[
         """
         EPCISDBProxy().get_full_message(
 
@@ -465,19 +478,25 @@ class EPCPyYesOutputStep(rules.Step, FilteredEventStepMixin):
         prepend_filtered_events = self.get_boolean_parameter(
             'Prepend Filtered Events', False
         )
+        only_filtered = self.get_boolean_parameter(
+            'Only Render Filtered Events', False
+        )
         oevents = rule_context.context.get(ContextKeys.OBJECT_EVENTS_KEY.value,
                                            [])
         aggevents = rule_context.context.get(
             ContextKeys.AGGREGATION_EVENTS_KEY.value, [])
-        if append_filtered_events:
-            if prepend_filtered_events:
-                all_events = self.get_filtered_events() + oevents + aggevents
-            else:
-                all_events = oevents + aggevents + self.get_filtered_events()
+        if only_filtered:
+            all_events = self.get_filtered_events()
         else:
-            all_events = oevents + aggevents
+            if append_filtered_events:
+                if prepend_filtered_events:
+                    all_events = self.get_filtered_events() + oevents + aggevents
+                else:
+                    all_events = oevents + aggevents + self.get_filtered_events()
+            else:
+                all_events = oevents + aggevents
         if len(all_events) > 0:
-            epcis_document = template_events.EPCISEventListDocument(all_events)
+            epcis_document = self.get_epcis_document_class(all_events)
             if self.get_boolean_parameter('JSON', False):
                 data = epcis_document.render_json()
             else:
@@ -486,6 +505,19 @@ class EPCPyYesOutputStep(rules.Step, FilteredEventStepMixin):
                 ContextKeys.OUTBOUND_EPCIS_MESSAGE_KEY.value
             ] = data
 
+    def get_epcis_document_class(self,
+                                 all_events) -> template_events.EPCISEventListDocument:
+        """
+        Override this to provide the step with an alternate class or to
+        override the default template provided.
+        :param all_events: The EPCPyYes template events to populate
+        the document with.
+        :return: By default returns a `template_events.EPCISEventListDocument`
+        instance.
+        """
+        epcis_document = template_events.EPCISEventListDocument(all_events)
+        return epcis_document
+
     def declared_parameters(self):
         return {
             "Append Filtered Events": _('Whether or not to append any events '
@@ -499,12 +531,17 @@ class EPCPyYesOutputStep(rules.Step, FilteredEventStepMixin):
                                          'added to the beginning of the EPCIS '
                                          'EventList- otherwise they will be '
                                          'added to the end.'),
+            "Only Render Filtered Events": _('If set to True, the step will '
+                                             'only render the filtered events. '
+                                             'If this is true, Append Filtered '
+                                             'Events and Prepend Filtered '
+                                             'Events will be ignored.'),
             "JSON": _('If set to True then the output message for the EPCPyYEs'
                       'events will be JSON.'),
         }
 
     def on_failure(self):
-        super().on_failure()
+        pass
 
 
 class CreateOutputTaskStep(rules.Step):
@@ -607,11 +644,26 @@ class CreateOutputTaskStep(rules.Step):
         pass
 
 
-class TransportStep(rules.Step, HttpTransportMixin, SftpTransportMixin):
+class TransportStep(rules.Step, HttpTransportMixin, SftpTransportMixin,
+                    MailMixin):
     '''
     Uses the transport information within the `EPCISOutputCriteria` placed
     on the context under the EPCIS_OUTPUT_CRITERIA_KEY to send any data that
     was placed on the context under the OUTBOUND_EPCIS_MESSAGE_KEY.
+
+    A transport rule is typically configured with one transport step and
+    the name of that rule would be specified in a different rule as a
+    step parameter to the CreateOutputTaskStep above for example.
+
+    Step Parameters:
+        'content-type': 'The content-type to add to the header during any '
+                        'http posts, puts, etc. Default is application/'
+                        'xml.'
+        'file-extension': 'The file extension to specify when posting and '
+                          'putting data via http. Default is xml.'
+        'body-raw': 'Whether or not the data should be sent as raw body '
+                    'or file attachment.'
+                    'Defaults to True.'
     '''
 
     def execute(self, data, rule_context: RuleContext):
@@ -699,6 +751,15 @@ class TransportStep(rules.Step, HttpTransportMixin, SftpTransportMixin):
                           output_criteria,
                           content_type,
                           file_extension)
+        elif protocol.lower() == 'mailto':
+            self.send_email(
+                data,
+                rule_context,
+                output_criteria,
+                self.info,
+                file_extension=self.get_parameter('file-extension', 'txt'),
+                mimetype=self.get_parameter('content-type', 'text/plain')
+            )
 
     def _supports_protocol(self, endpoint: EndPoint):
         '''
@@ -712,7 +773,7 @@ class TransportStep(rules.Step, HttpTransportMixin, SftpTransportMixin):
         parse_result = urlparse(
             endpoint.urn
         )
-        if parse_result.scheme.lower() in ['http', 'https', 'sftp']:
+        if parse_result.scheme.lower() in ['http', 'https', 'sftp', 'mailto']:
             return parse_result.scheme
         else:
             raise errors.ProtocolNotSupportedError(_(
@@ -731,6 +792,145 @@ class TransportStep(rules.Step, HttpTransportMixin, SftpTransportMixin):
                             'xml',
             'file-extension': 'The file extension to specify when posting and '
                               'putting data via http. Default is xml',
-            'body-raw': 'Whether or not the data should be sent as raw body or file attachment.'
+            'body-raw': 'Whether or not the data should be sent as raw body '
+                        'or file attachment.'
                         'Defaults to True.',
         }
+
+
+class DelayStep(rules.Step):
+    """
+    Will simply introduce a delay (in seconds) within a rule.  This
+    step has a default delay setting of one second but can be configured
+    to allow uo to ten seconds of delay via a Timeout Interval step
+    parameter.
+    """
+
+    def execute(self, data, rule_context: RuleContext):
+        self.info('Checking for the Timeout Interval step parameter. '
+                  'The default is 1 second.  The parameter value is '
+                  'in seconds with a max of 10.')
+        timeout_interval = self.get_parameter(
+            parameter_name='Timeout Interval', default=0
+        )
+        # if it comes from the database it will be a string
+        timeout_interval = int(timeout_interval)
+        self.info(
+            'Sleeping the thread for %s seconds...' % timeout_interval)
+        if timeout_interval > 0:
+            time.sleep(timeout_interval)
+
+    @property
+    def declared_parameters(self):
+        return {
+            'Timeout Interval': 'The amount of time in seconds to pause the '
+                                'rule.'
+        }
+
+    def on_failure(self):
+        pass
+
+
+class EPCPyYesFilteredEventOutputStep(rules.Step, FilteredEventStepMixin):
+    """
+    Very similar to the EPCPyYesOutput step except that this step will
+    only render the filtered events and place them into the same outbound
+    OUTBOUND_EPCIS_MESSAGE_KEY context key for processing by a
+    CreateOutputTaskStep later in the rule for example.
+    """
+
+    def execute(self, data, rule_context: RuleContext):
+        """
+        Will pull any filtered events off of the rule context using the
+        FILTERED_EVENTS_KEY context key.
+        :param data: The data coming into the step from the rule.
+        :param rule_context: A reference to the rule context.
+        """
+        self.rule_context = rule_context
+        filtered_events = self.get_filtered_events()
+        self.info('Found %s filtered events.' % len(filtered_events))
+        if len(filtered_events) >= 0:
+            epcis_document = template_events.EPCISEventListDocument(
+                filtered_events)
+            if self.get_boolean_parameter('JSON', False):
+                data = epcis_document.render_json()
+            else:
+                data = epcis_document.render()
+            self.info('Warning: this step is overwriting the Outbound '
+                      'EPCIS Message key context key data.  If any data '
+                      'was in this key prior to this step and had not '
+                      'yet been processed, it will have been overwritten.')
+            rule_context.context[
+                ContextKeys.OUTBOUND_EPCIS_MESSAGE_KEY.value
+            ] = data
+
+    @property
+    def declared_parameters(self):
+        return {}
+
+    def on_failure(self):
+        pass
+
+
+class AppendCommissioningStep(rules.Step, FilteredEventStepMixin):
+    """
+    Will take all of the unique epcs found in the filtered events
+    and create an EPCPyYes commissioning event for those epcs and append it to
+    the RuleContext's OBJECT_EVENTS list.  This is different than the
+    `AddCommissioningStep above which will look up all of the commissioning
+    events of the epcis in the message along with the children of the epcs
+    within the message.  This step does not do any recurssive child lookups
+    so it is much more efficient to use if you are intending to just lookup
+    the commissioning events for the EPCs in a given message.
+    """
+
+    def execute(self, data, rule_context: RuleContext):
+        epcs = set({})
+        for filtered_event in rule_context.context.get(
+            ContextKeys.FILTERED_EVENTS_KEY.value,
+            []):
+            if isinstance(filtered_event, template_events.AggregationEvent):
+                epcs = epcs.union(filtered_event.child_epcs)
+                epcs.add(filtered_event.parent_id)
+            elif isinstance(filtered_event, template_events.ObjectEvent):
+                epcs = epcs.union(filtered_event.epc_list)
+            elif isinstance(filtered_event, template_events.TransactionEvent):
+                epcs = epcs.union(filtered_event.epc_list)
+                epcs.add(filtered_event.parent_id)
+            elif isinstance(filtered_event,
+                            template_events.TransformationEvent):
+                raise self.TransformationEventNotSupported(
+                    'Auto commissioning for transformation events is not '
+                    'supported.'
+                )
+            else:
+                raise self.EventNotSupported(
+                    'An event was passed in that was not supported. '
+                    'Only EPCPyYes template events can be processed '
+                    'by this step.'
+                )
+
+        obj_events = self.get_object_events(list(epcs))
+        rule_context.context[
+            ContextKeys.OBJECT_EVENTS_KEY.value
+        ] = obj_events
+
+    def get_object_events(self, epcs):
+        db_proxy = EPCISDBProxy()
+        obj_events = db_proxy.get_object_events_by_epcs(
+            epcs,
+            select_for_update=False
+        )
+        return obj_events
+
+    def declared_parameters(self):
+        pass
+
+    def on_failure(self):
+        pass
+
+    class TransformationEventNotSupported(Exception):
+        pass
+
+    class EventNotSupported(Exception):
+        pass
