@@ -8,20 +8,25 @@ test_quartet_output
 Tests for `quartet_output` models module.
 """
 import os
+from urllib.parse import urlparse
+
 import paramiko
+import time
+from django.conf import settings
 from django.core.exceptions import ValidationError
-from EPCPyYes.core.v1_2.events import EventType
-from EPCPyYes.core.v1_2.CBV.dispositions import Disposition
-from EPCPyYes.core.v1_2.CBV.business_steps import BusinessSteps
-from EPCPyYes.core.SBDH.sbdh import StandardBusinessDocumentHeader as sbdheader
-from quartet_epcis.parsing.business_parser import BusinessEPCISParser
-from quartet_capture.models import Rule, Step, StepParameter, Task
-from quartet_capture.tasks import execute_rule, execute_queued_task
-from quartet_output.steps import SimpleOutputParser, ContextKeys
-from quartet_output.models import EPCISOutputCriteria
 from django.test import TestCase
 
+from EPCPyYes.core.SBDH.sbdh import StandardBusinessDocumentHeader as sbdheader
+from EPCPyYes.core.v1_2.CBV.business_steps import BusinessSteps
+from EPCPyYes.core.v1_2.CBV.dispositions import Disposition
+from EPCPyYes.core.v1_2.events import EventType
+from quartet_capture.models import Rule, Step, StepParameter, Task
+from quartet_capture.tasks import execute_rule, execute_queued_task
+from quartet_epcis.parsing.business_parser import BusinessEPCISParser
 from quartet_output import models
+from quartet_output.models import EPCISOutputCriteria
+from quartet_output.steps import SimpleOutputParser, ContextKeys
+from quartet_output.transport.mail import MailMixin
 
 
 class TestQuartetOutput(TestCase):
@@ -100,6 +105,53 @@ class TestQuartetOutput(TestCase):
                     ContextKeys.EPCIS_OUTPUT_CRITERIA_KEY.value)
             )
 
+    def test_rule_with_agg_trigger(self):
+        self._create_good_agg_trigger_ouput_criterion()
+        db_rule = self._create_rule()
+        self._create_step(db_rule)
+        self._create_output_steps(db_rule)
+        self._create_comm_step(db_rule)
+        self._create_epcpyyes_step(db_rule)
+        db_task = self._create_task(db_rule)
+        curpath = os.path.dirname(__file__)
+        self._parse_test_data('data/commissioning.xml')
+        self._parse_test_data('data/aggregation.xml')
+
+    def test_rule_with_agg_comm_json_output(self):
+        self._create_good_ouput_criterion()
+        db_rule = self._create_rule()
+        self._create_step(db_rule)
+        self._create_output_steps(db_rule)
+        self._create_comm_step(db_rule)
+        self._create_epcpyyes_step(db_rule, json=True)
+        self._create_task_step(db_rule)
+        db_rule2 = self._create_transport_rule()
+        self._create_transport_step(db_rule2)
+        db_task = self._create_task(db_rule)
+        curpath = os.path.dirname(__file__)
+        # prepopulate the db
+        self._parse_test_data('data/commission_one_event.xml')
+        self._parse_test_data('data/nested_pack.xml')
+        data_path = os.path.join(curpath, 'data/ship_pallet.xml')
+        with open(data_path, 'r') as data_file:
+            context = execute_rule(data_file.read().encode(), db_task)
+            self.assertEqual(
+                len(context.context[ContextKeys.AGGREGATION_EVENTS_KEY.value]),
+                3,
+                "There should be three filtered events."
+            )
+            for event in context.context[
+                ContextKeys.AGGREGATION_EVENTS_KEY.value]:
+                if event.parent_id in ['urn:epc:id:sgtin:305555.3555555.1',
+                                       'urn:epc:id:sgtin:305555.3555555.2']:
+                    self.assertEqual(len(event.child_epcs), 5)
+                else:
+                    self.assertEqual(len(event.child_epcs), 2)
+            task_name = context.context[ContextKeys.CREATED_TASK_NAME_KEY]
+            execute_queued_task(task_name=task_name)
+            task = Task.objects.get(name=task_name)
+            self.assertEqual(task.status, 'FINISHED')
+
     def test_rule_with_agg_comm_output(self):
         self._create_good_ouput_criterion()
         db_rule = self._create_rule()
@@ -108,7 +160,44 @@ class TestQuartetOutput(TestCase):
         self._create_comm_step(db_rule)
         self._create_epcpyyes_step(db_rule)
         self._create_task_step(db_rule)
-        db_rule2= self._create_transport_rule()
+        db_rule2 = self._create_transport_rule()
+        self._create_transport_step(db_rule2)
+        db_task = self._create_task(db_rule)
+        curpath = os.path.dirname(__file__)
+        # prepopulate the db
+        self._parse_test_data('data/commission_one_event.xml')
+        self._parse_test_data('data/nested_pack.xml')
+        data_path = os.path.join(curpath, 'data/ship_pallet.xml')
+        with open(data_path, 'r') as data_file:
+            context = execute_rule(data_file.read().encode(), db_task)
+            self.assertEqual(
+                len(context.context[ContextKeys.AGGREGATION_EVENTS_KEY.value]),
+                3,
+                "There should be three filtered events."
+            )
+            for event in context.context[
+                ContextKeys.AGGREGATION_EVENTS_KEY.value]:
+                if event.parent_id in ['urn:epc:id:sgtin:305555.3555555.1',
+                                       'urn:epc:id:sgtin:305555.3555555.2']:
+                    self.assertEqual(len(event.child_epcs), 5)
+                else:
+                    self.assertEqual(len(event.child_epcs), 2)
+            task_name = context.context[ContextKeys.CREATED_TASK_NAME_KEY]
+            execute_queued_task(task_name=task_name)
+            task = Task.objects.get(name=task_name)
+            self.assertEqual(task.status, 'FINISHED')
+
+    def test_rule_with_agg_comm_output_delay(self):
+        self._create_good_ouput_criterion()
+        db_rule = self._create_rule()
+        self._create_step(db_rule)
+        self._create_output_steps(db_rule)
+        self._create_comm_step(db_rule)
+        self._create_epcpyyes_step(db_rule)
+        self._create_task_step(db_rule)
+        self.create_delay_step(db_rule, order=10)
+        self._create_filtered_output_step(db_rule, order=20)
+        db_rule2 = self._create_transport_rule()
         self._create_transport_step(db_rule2)
         db_task = self._create_task(db_rule)
         curpath = os.path.dirname(__file__)
@@ -169,7 +258,7 @@ class TestQuartetOutput(TestCase):
         self._create_comm_step(db_rule)
         self._create_epcpyyes_step(db_rule)
         self._create_task_step(db_rule)
-        db_rule2= self._create_transport_rule()
+        db_rule2 = self._create_transport_rule()
         self._create_transport_step(db_rule2, put_data=True)
         db_task = self._create_task(db_rule)
         curpath = os.path.dirname(__file__)
@@ -275,6 +364,21 @@ class TestQuartetOutput(TestCase):
             except EPCISOutputCriteria.DoesNotExist:
                 pass
 
+    def test_delay_rule(self):
+        start = time.time()
+        db_rule = self._create_delay_rule()
+        db_task = self._create_task(db_rule)
+        curpath = os.path.dirname(__file__)
+        data_path = os.path.join(curpath, 'data/epcis.xml')
+        with open(data_path, 'r') as data_file:
+            try:
+                execute_rule(data_file.read().encode(), db_task)
+            except EPCISOutputCriteria.DoesNotExist:
+                pass
+        stop = time.time()
+        self.assertGreaterEqual((stop - start), 3, 'The rule did not pause '
+                                                   'long enough.')
+
     def _create_good_ouput_criterion(self):
         endpoint = self._create_endpoint()
         auth = self._create_auth()
@@ -295,6 +399,19 @@ class TestQuartetOutput(TestCase):
         eoc.save()
         return eoc
 
+    def _create_good_agg_trigger_ouput_criterion(self):
+        endpoint = self._create_endpoint()
+        auth = self._create_auth()
+        eoc = EPCISOutputCriteria()
+        eoc.name = "Test Criteria"
+        eoc.action = "ADD"
+        eoc.biz_step = BusinessSteps.packing.value
+        eoc.biz_location = 'urn:epc:id:sgln:0555555.00002.0'
+        eoc.sender_identifier = 'urn:epc:id:sgln:0555555.00001.0'
+        eoc.end_point = endpoint
+        eoc.save()
+        return eoc
+
     def _create_good_header_criterion(self):
         endpoint = self._create_endpoint()
         auth = self._create_auth()
@@ -351,6 +468,27 @@ class TestQuartetOutput(TestCase):
         parser.parse()
         parser.clear_cache()
 
+    def _create_delay_rule(self):
+        rule = Rule()
+        rule.name = 'delay-rule'
+        rule.description = 'a simple delay rule'
+        rule.save()
+        self.create_delay_step(rule)
+        return rule
+
+    def create_delay_step(self, rule, order=1):
+        step = Step()
+        step.step_class = 'quartet_output.steps.DelayStep'
+        step.order = order
+        step.name = 'wait 3 seconds'
+        step.rule = rule
+        step.save()
+        param = StepParameter()
+        param.step = step
+        param.name = 'Timeout Interval'
+        param.value = '3'
+        param.save()
+
     def _create_rule(self):
         rule = Rule()
         rule.name = 'output-test'
@@ -358,6 +496,14 @@ class TestQuartetOutput(TestCase):
         rule.save()
         return rule
 
+    def _create_filtered_output_step(self, rule, order=1):
+        step = Step()
+        step.step_class = 'quartet_output.steps.EPCPyYesFilteredEventOutputStep'
+        step.order = order
+        step.name = 'filtered output step'
+        step.rule = rule
+        step.save()
+
     def _create_transport_rule(self):
         rule = Rule()
         rule.name = 'Transport Rule'
@@ -417,7 +563,7 @@ class TestQuartetOutput(TestCase):
         step.description = 'unit test commissioning step'
         step.save()
 
-    def _create_epcpyyes_step(self, rule):
+    def _create_epcpyyes_step(self, rule, json=False):
         step = Step()
         step.rule = rule
         step.order = 4
@@ -426,6 +572,12 @@ class TestQuartetOutput(TestCase):
         step.description = 'Creates EPCIS XML or JSON and inserts into rule' \
                            'context.'
         step.save()
+        if json:
+            param = StepParameter.objects.create(
+                step=step,
+                name='JSON',
+                value=True
+            )
 
     def _create_task(self, rule):
         task = Task()
@@ -444,10 +596,10 @@ class TestQuartetOutput(TestCase):
                                      'rule.'
         step_parameter.save()
 
-    def _create_task_step(self, rule):
+    def _create_task_step(self, rule, order=5):
         step = Step()
         step.rule = rule
-        step.order = 5
+        step.order = order
         step.name = 'Create Output Task'
         step.step_class = 'quartet_output.steps.CreateOutputTaskStep'
         step.description = 'Looks for any EPCIS data on the context and ' \
@@ -492,7 +644,6 @@ class TestQuartetOutput(TestCase):
         pass
 
 
-    
 class TestSFTPTransport(TestQuartetOutput):
 
     def _create_endpoint(self):
@@ -538,13 +689,24 @@ class TestSFTPTransport(TestQuartetOutput):
             self.assertEqual(task.status, 'FINISHED')
             try:
                 sftp_client = paramiko.SSHClient()
-                sftp_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
-                sftp_client.connect('testsftphost', '22', username='foo', password='pass', timeout=60)
+                sftp_client.set_missing_host_key_policy(
+                    paramiko.AutoAddPolicy())
+                sftp_client.connect('testsftphost', '22', username='foo',
+                                    password='pass', timeout=60)
                 sftp = sftp_client.open_sftp()
-                remote_file = sftp.open('/upload/'+task_name+'.xml', 'r')
+                remote_file = sftp.open('/upload/' + task_name + '.xml', 'r')
                 data_back = remote_file.read()
                 self.assertEqual(data_back.decode(), data_out)
             finally:
                 sftp_client.close()
-                
-                
+
+class TestMail(TestCase):
+    def test_mailto(self):
+        urn = 'mailto:hither@this.local?body=send%20current-issue&subject=awesome email'
+        mail_info = urlparse(urn)
+        mixin = MailMixin()
+        print(settings.EMAIL_BACKEND)
+        message = mixin.convert_mailto_url(urn)
+        message.attach('test.txt', '<test>asdf</test>', 'text/plain')
+        message.send(fail_silently=False)
+        print(mail_info)
